/****************************************************************************
**  CUBE        http://www.score-p.org/                                    **
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 2011-2014                                                **
**  Forschungszentrum Juelich GmbH, Juelich Supercomputing Centre          **
**                                                                         **
**  Copyright (c) 2011-2014                                                **
**  German Research School for Simulation Sciences GmbH,                   **
**  Laboratory for Parallel Programming                                    **
**                                                                         **
**  This software may be modified and distributed under the terms of       **
**  a BSD-style license.  See the COPYING file in the package base         **
**  directory for details.                                                 **
****************************************************************************/

/**
* Cube model.
*/
package scalasca.cubex.cube;
import java.lang.Integer;
import java.lang.Long;
import java.lang.String;

import java.util.Map;
import java.util.HashMap;
import java.util.ArrayList;
import java.io.File;
import java.io.FileInputStream;
import java.io.ByteArrayInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import javax.swing.filechooser.FileFilter;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

import scalasca.cubex.cubexlayout.*;
import scalasca.cubex.cube.cubeparser.*;
import scalasca.cubex.cube.datalayout.data.value.*;
import scalasca.cubex.cube.errors.*;
import scalasca.cubex.cube.services.*;


/*
 * Class , delivers file structure of CubeX
 * 
 */
public class Cube  extends ProgressReport
{
// ---------------- PRIVATE -----------------------------------
	private boolean loading = false;

	private FileInputStream fis;
	private BufferedInputStream buffered_fis;

	private CubeLayout cube_tar_layout;

// --------------- PROTECTED  -------------------------------

	protected String filename;

	protected String version;
	

	protected Map<String, String> attributes = new HashMap< String, String > ();
	protected ArrayList<String> mirrors = new ArrayList< String> ();

	protected ArrayList<Metric> metv = new ArrayList<Metric>();
	protected ArrayList<Metric> root_metv = new ArrayList<Metric>();

	protected ArrayList<Region> regionv = new ArrayList<Region>();
	protected ArrayList<Cnode> cnodev = new ArrayList<Cnode>();
	protected ArrayList<Cnode> root_cnodev = new ArrayList<Cnode>();

/*
	protected ArrayList<Machine> machinev = new ArrayList<Machine>();
	protected ArrayList<Node> nodev = new ArrayList<Node>();
	protected ArrayList<scalasca.cubex.cube.Process> processv = new ArrayList<scalasca.cubex.cube.Process>();
	protected ArrayList<scalasca.cubex.cube.Thread> thrdv = new ArrayList<scalasca.cubex.cube.Thread>();*/


	protected ArrayList<SystemTreeNode> stnv = new ArrayList<SystemTreeNode>();
	protected ArrayList<SystemTreeNode> root_stnv = new ArrayList<SystemTreeNode>();
	protected ArrayList<SystemTreeNode> non_root_stnv = new ArrayList<SystemTreeNode>();

	protected ArrayList<LocationGroup> lgv = new ArrayList<LocationGroup>();
	protected ArrayList<Location> locv = new ArrayList<Location>();	
	
	
	
	protected ArrayList<Cartesian> cartv = new ArrayList<Cartesian>();


	protected String metricsTreeTitle = "";
	protected String callTreeTitle = "";
	protected String systemTreeTitle = "";



	public Cube() {super(); }

	public void openCubeReport(String _filename) throws BadSyntaxException, BadCubeReportLayoutException,  FileNotFoundException, IOException
	{
		try
		{
			progressContext.setNextStep(0.04);
			progressContext.setProgressMessage("Open report and get the achor file...");
			loading = true;
			filename = _filename;
			File file= new File(filename);

			progressContext.setAchievedStep();
			progressContext.setNextStep(0.22);
			progressContext.setProgressMessage("Parsing report structure...");
			cube_tar_layout =  CubeLayout.getCubeLayout(file.getPath()); // detects and returns either tar layout for cubex or faked layout for cube3;

			progressContext.setAchievedStep();
			progressContext.setNextStep(0.04);
			progressContext.setProgressMessage("Parsing meta data of the report...");

			byte anchor[] = cube_tar_layout.getAnchorBytes();
			ByteArrayInputStream bis = new ByteArrayInputStream(anchor);

			XMLReader xmlreader = XMLReaderFactory.createXMLReader("org.apache.xerces.parsers.SAXParser");
			CubeParser cubeParser = new CubeParser(this);
			xmlreader.setContentHandler(cubeParser);
			xmlreader.setErrorHandler(cubeParser);
			progressContext.setSubActor(cubeParser);
			xmlreader.parse(new InputSource(bis));
			progressContext.setAchievedStep();
			progressContext.setProgressMessage("Load binary data...");
			loadData();
			cube_tar_layout=null; // release memory
			System.gc();          // several calls are usually needed to perform real memory release
			System.gc();
			System.gc();
			System.gc();
			loading = false;
		}
		catch (BadCubeReportLayoutException e) {
			throw e;
		}
		catch (SAXException e) {
			throw new BadSyntaxException(e.getMessage());
		}
	}


// --------------- cubewide methods -----------------
	public void set_version (String _version)
	{
		version = _version;
	}
	public String get_version ()
	{
		return version;
	}



	public void def_attr (String key, String value)
	{
		if (key != null)
		{
			attributes.put(key, value);
		}
	}
	
	
    public String get_attr (String key)
    {
        if (key != null)
        {
            return attributes.get(key);
        }else 
        return null;
    }
	

	public void def_mirror (String mirror)
	{
		if (mirror != null)
		{
			mirrors.add(mirror);
		}
	}

//  ------------- dimensions methods --------------------

 	public Metric def_met(String uniq_name, String disp_name, String dtype, String uom, String url, String descr, MetricType metricType, Metric parent)
	throws UnknownMetricTypeException, UnknownValueTypeException
	{
		int _id = metv.size();
		Metric _met = def_met(_id, metricType, parent);
		if (_met != null) 
		{	
			_met.init(	uniq_name,
					disp_name,
					dtype,
					uom,
					url,
					descr);
		}
		return _met;
	}


 	public Metric def_met(int id, String uniq_name, String disp_name, String dtype, String uom, String url, String descr, MetricType metricType, Metric parent)
	throws UnknownMetricTypeException, UnknownValueTypeException
	{
		Metric _met = def_met(id, metricType, parent);
		if (_met != null)
		{
			_met.init(	uniq_name,
					disp_name,
					dtype,
					uom,
					url,
					descr);
		}
		return _met;
	}


	public Metric def_met(int id, MetricType metricType, Metric parent)
		throws UnknownMetricTypeException
	{
		Metric _met = null;
		if (metricType != MetricType.DERIVED ) // we ignore now derived metrics.... not supported yet
		{
		    if (metv.size()<=id)
		    {
			    metv.ensureCapacity(id+1);
			    for (int i= metv.size(); i< (id+1); i++)
				    metv.add(null);
		    }
		    switch (metricType)
		    {
			    case INCLUSIVE:
				    _met = new InclusiveMetric(id, parent);
				    metv.set(id, _met);
				    if (parent == null)
					    root_metv.add(_met);
				    break;
			    case EXCLUSIVE:
				    _met = new ExclusiveMetric(id,  parent);
				    metv.set(id, _met);
				    if (parent == null)
					    root_metv.add(_met);
				    break;
			    case DERIVED: // we ignore now derived metrics.... not supported yet
				    break;
			    default:
				    throw new UnknownMetricTypeException(metricType);
		    }
		}
		return _met;
	}





	public Region def_region(String name, int beginln, int endln, String url, String descr, String mod)
	{
		int _id = regionv.size();
		return def_region( name, beginln, endln, url, descr, mod);
	}

	public Region def_region(int id, String name, int beginln, int endln, String url, String descr, String mod)
	{
		Region _reg = null;
		if (regionv.size()<=id)
		{
			regionv.ensureCapacity(id+1);
			for (int i= regionv.size(); i< (id+1); i++)
				regionv.add(null);
		}
		_reg = new Region(id, name, beginln, endln, mod,  url, descr);
		regionv.set(id, _reg);
		return _reg;
	}




	public Cnode def_cnode(int region_id, Cnode parent)
	{
		int _id = cnodev.size();
		return def_cnode(_id, region_id, parent);
	}

	public Cnode def_cnode(int id, int region_id, Cnode parent)
	{
		if (cnodev.size()<=id)
		{
			cnodev.ensureCapacity(id+1);
			for (int i= cnodev.size(); i< (id+1); i++)
				cnodev.add(null);
		}
		Region _reg = regionv.get(region_id);
		Cnode _cnode = new Cnode(id, _reg, parent);
		cnodev.set(id, _cnode);
		if (parent == null)
			root_cnodev.add(_cnode);
		return _cnode;
	}




	public SystemTreeNode def_system_tree_node(String name, String descr, String _class, SystemTreeNode parent)
	{
		int _id =stnv.size();
		return def_system_tree_node(_id, name, descr, _class, parent);
	}

	public SystemTreeNode def_system_tree_node(int id, SystemTreeNode parent)
	{
		if (stnv.size()<=id)
		{
			stnv.ensureCapacity(id+1);
			for (int i= stnv.size(); i< (id+1); i++)
				stnv.add(null);
		}
		SystemTreeNode _stn = new SystemTreeNode(id, parent);
		stnv.set(id, _stn);
		if (parent == null)
		  root_stnv.add(_stn);
		else
		  non_root_stnv.add(_stn);
		return _stn;
	}


	public SystemTreeNode def_system_tree_node(int id, String name, String descr, String _class, SystemTreeNode parent)
	{
		SystemTreeNode _stn = def_system_tree_node(id, parent);
		_stn.init(name, descr, _class);
		return _stn;
	}


	
	

	public SystemTreeNode def_machine(String name, String descr)
	{
	      return new Machine(def_system_tree_node(name, descr, "machine",  (SystemTreeNode)null));
	}

	public SystemTreeNode def_machine(int id)
	{
	      return new Machine(def_system_tree_node(id, "noname machine", "", "machine", (SystemTreeNode)null));
	}


	public SystemTreeNode def_machine(int id, String name, String descr)
	{
	      return new Machine(def_system_tree_node(id, name, descr, "machine", (SystemTreeNode)null));
	}


	public SystemTreeNode def_node(String name, String descr, SystemTreeNode parent)
	{
	      return new Node(def_system_tree_node(name, descr, "node", parent), parent);
	}

	public SystemTreeNode def_node(int id, SystemTreeNode parent)
	{
	      return new Node(def_system_tree_node(id, "noname node", "", "node", parent), parent);
	}


	public SystemTreeNode def_node(int id, String name, String descr, SystemTreeNode parent)
	{
	      return new Node(def_system_tree_node(id, name, descr, "node", parent), parent);
	}



	public LocationGroup def_location_group(long rank, String name, LocationGroupType type, SystemTreeNode parent)
	{
		int _id =lgv.size();
		LocationGroup _lg = def_location_group(_id, parent);
		_lg.init(rank, name, type);
		return _lg;
	}

	public LocationGroup def_location_group(int id, SystemTreeNode parent)
	{
		if (lgv.size()<=id)
		{
			lgv.ensureCapacity(id+1);
			for (int i= lgv.size(); i< (id+1); i++)
				lgv.add(null);
		}
		LocationGroup _lg = new LocationGroup(id, parent);
		parent.addLocationGroup( _lg );
		lgv.set(id, _lg);
		return _lg;
	}


	public LocationGroup def_location_group(int id, long rank, String name, LocationGroupType type,  SystemTreeNode parent)
	{
		LocationGroup _lg = def_location_group(id, parent);
		_lg.init(rank, name, type);
		return _lg;
	}	
	
	
	public LocationGroup def_proc(long rank, String name,SystemTreeNode parent)
	{
		return new Process(def_location_group(rank, name, LocationGroupType.PROCESS, parent), parent);
	}

	public LocationGroup def_proc(int id, SystemTreeNode parent)
	{
		return new Process(def_location_group(id, parent), parent);
	}


	public LocationGroup def_proc(int id, long rank,  String name, SystemTreeNode parent)
	{
		return new Process(def_location_group(id, rank, name, LocationGroupType.PROCESS, parent), parent);
	}




	public Location def_location(long rank, String name, LocationType type, LocationGroup parent)
	{
		int _id =locv.size();
		return def_location(_id, rank, name, type,  parent);
	}

	public Location def_location(int id, LocationGroup parent)
	{
		if (locv.size()<=id)
		{
			locv.ensureCapacity(id+1);
			for (int i= locv.size(); i< (id+1); i++)
				locv.add(null);
		}
		Location _thrd = new Location(id, parent);
		_thrd.init(0,"",LocationType.CPU_THREAD);
		locv.set(id, _thrd);
		return _thrd;
	}


	public Location def_location(int id, long rank,  String name, LocationType type, LocationGroup parent)
	{
		Location _thread = def_thrd(id, parent);
		_thread.init(rank, name, type);
		return _thread;
	}
 
 
 
 
	public Location def_thrd(long rank, String name,  LocationGroup parent)
	{
		return new scalasca.cubex.cube.Thread(def_location(rank, name,LocationType.CPU_THREAD, parent), parent);
	}

	public Location def_thrd(int id, LocationGroup parent)
	{
		return new scalasca.cubex.cube.Thread(def_location(id, parent), parent);
	}

	public Location def_thrd(int id, long rank,  String name, LocationGroup parent)
	{
		return new scalasca.cubex.cube.Thread(def_location(id, rank, name,LocationType.CPU_THREAD, parent), parent);
	}


	public Cartesian def_cart( int ndim, ArrayList<Integer> dimv, ArrayList<Boolean> periodv)
	{
		Cartesian _cart = new Cartesian(ndim, dimv, periodv);
		cartv.add(_cart);
		return _cart;
	}


// 	public void def_coods(Cartesian cart, scalasca.cubex.cube.Thread thread, ArrayList<Integer> coordv)
// 	{
// 		cart.def_coords(thread, coordv);
// 	}



// ---------------- return dimenstion methods ---------------------------


	public ArrayList<Metric> get_metv() { return metv;}
	public ArrayList<Metric> get_root_metv() { return root_metv;}

	public ArrayList<Region> get_regionv() { return regionv;}
	public ArrayList<Cnode> get_cnodev() { return cnodev;}
	public ArrayList<Cnode> get_root_cnodev() { return root_cnodev;}

    public ArrayList<SystemTreeNode> get_root_stnv() { return root_stnv; } 

	
	public ArrayList<Machine> get_machv() 
	{ 
	  ArrayList<Machine> _to_return= new ArrayList<Machine>();
	  for (SystemTreeNode _obj: stnv )
	  {
        if (_obj.getSystemTreeNodeClass() == "machine")
            _to_return.add((scalasca.cubex.cube.Machine)(_obj));
	  }
	  return _to_return;
	}
	
	public ArrayList<Node> get_nodev() 	
	{ 
	  ArrayList<scalasca.cubex.cube.Node> _to_return= new ArrayList<scalasca.cubex.cube.Node>();
	  for (SystemTreeNode _obj: stnv )
	  {
        if (_obj.getSystemTreeNodeClass() == "node")
            _to_return.add((scalasca.cubex.cube.Node)_obj);
      }
	  return _to_return;
	}
	
	public ArrayList<Process> get_procv() 	
	{ 
	  ArrayList<scalasca.cubex.cube.Process> _to_return= new ArrayList<scalasca.cubex.cube.Process>();
	  for (scalasca.cubex.cube.LocationGroup _obj: lgv )
	  {
            if (_obj.getType() == LocationGroupType.PROCESS)
                _to_return.add((scalasca.cubex.cube.Process)_obj);
	  }
	  return _to_return;
	}
	
	public ArrayList<scalasca.cubex.cube.Thread>      get_thrdv() 	
	{ 
	  ArrayList<scalasca.cubex.cube.Thread> _to_return= new ArrayList<scalasca.cubex.cube.Thread>();
	  for (scalasca.cubex.cube.Location _obj: locv )
	  {
        if (_obj.getType() == LocationType.CPU_THREAD)
            _to_return.add((scalasca.cubex.cube.Thread)_obj);
	  }
	  return _to_return;
	}

	public ArrayList<SystemTreeNode> get_stnv() { return stnv;}
	public ArrayList<LocationGroup> get_location_groupv() { return lgv;}
	public ArrayList<Location> get_locationv() { return locv;}	
	
	public ArrayList<Cartesian> get_cartv() { return cartv;}

// ----------------- set titles methods ------------------------

	public void set_metricstree_title(String title) { metricsTreeTitle = title; } 
	public void set_calltree_title(String title)    { callTreeTitle = title; }
	public void set_systemtree_title(String title)  { systemTreeTitle = title; }

	public String get_metricstree_title() { return metricsTreeTitle; }
	public String get_calltree_title()    { return callTreeTitle; }
	public String get_systemtree_title()  { return systemTreeTitle; }


// ------------------ loading progress method  --------------------------------


	public boolean isLoading()
	{
		return loading;
	}


 // ------------------- Set severity calls -------------------------------------
	public void set_sev(Metric met, Cnode cnode, Location loc, String value) throws NotEnumeratedCnodeException
	{
		met.set_sev( cnode, loc, value );
	}

	public void set_sev(Metric met, Cnode cnode, Thread th, String value) throws NotEnumeratedCnodeException
	{
		met.set_sev( cnode, (Location)th, value );
	}





 // ------------------ Get severity calls --------------------------------------


	public double get_sev(Metric met, Cnode cnode, Location loc) throws NotEnumeratedCnodeException
	{
		return get_sev(met, cnode, CubeCalculationFlavour.EXCLUSIVE, loc );
	}
	public double get_sev(Metric met, Cnode cnode, scalasca.cubex.cube.Thread thread) throws NotEnumeratedCnodeException
	{
		return get_sev(met, cnode, CubeCalculationFlavour.EXCLUSIVE, (Location)thread );
	}


	public double get_sev(Metric met, Cnode cnode, CubeCalculationFlavour cf, Location loc) throws NotEnumeratedCnodeException
	{
		Value _v = get_sev_adv(met, cnode, cf, loc);
		return _v.getDouble();
	}
	public double get_sev(Metric met, Cnode cnode, CubeCalculationFlavour cf, scalasca.cubex.cube.Thread thread) throws NotEnumeratedCnodeException
	{

		Value _v = get_sev_adv(met, cnode, cf, (Location)thread);
		return _v.getDouble();
	}
	
	public Value get_sev_adv(Metric met, Cnode cnode, CubeCalculationFlavour cf, Location loc) throws NotEnumeratedCnodeException
	{
		return met.get_sev(cnode, cf, loc);
	}
	
	public Value get_sev_adv(Metric met, Cnode cnode, CubeCalculationFlavour cf, scalasca.cubex.cube.Thread  thread) throws NotEnumeratedCnodeException
	{
		return met.get_sev(cnode, cf, (Location)thread);
	}

	public Value get_sev_adv(Metric met, Cnode cnode, Location loc) throws NotEnumeratedCnodeException
	{
		return get_sev_adv(met, cnode, CubeCalculationFlavour.EXCLUSIVE, loc);
	}
	public Value get_sev_adv(Metric met, Cnode cnode, scalasca.cubex.cube.Thread thread) throws NotEnumeratedCnodeException
	{
		return get_sev_adv(met, cnode, CubeCalculationFlavour.EXCLUSIVE, (Location)thread);
	}






// ------------------------ LOADING DATA and CREATING all data Structures ------------------------
	private void loadData() throws BadCubeReportLayoutException
	{
		if (! 
            (
              version.equalsIgnoreCase("4.0")
              ||
              version.equalsIgnoreCase("4.1")
              ||
              version.equalsIgnoreCase("4.2")
              ||
              version.equalsIgnoreCase("4.3")
            ))
		{
			progressContext.setNextStep(0.30);
			return; // in not 4th version (3th) no binary data.
		}

		double step = 0.7/metv.size();

		for (scalasca.cubex.cube.Metric met : metv)
		{
			progressContext.setNextStep(step);
			progressContext.setProgressMessage("Load data into CUBE for metric: "+met.getDisplayName());
			met.createEnumeration(root_cnodev);
            try
			{
				byte metric_index[] = cube_tar_layout.getMetricIndexBytes( met.getId());
				byte metric_data[] =  cube_tar_layout.getMetricDataBytes(  met.getId());

				// deliver to the metric filled with content of metrc's files
				progressContext.setSubActor(met);

				if (metric_index != null  && metric_data != null)
				  met.createDataMatrix(metric_index, metric_data, cnodev.size(), locv.size());
				else
				  met.createDataMatrix(cnodev.size(), locv.size());
				progressContext.setAchievedStep();
				metric_index=null;  // release memory
				metric_data=null;   // release memory
			}
			catch (Exception e)
			{
					throw new BadCubeReportLayoutException("Cannot load data for metric "+met.getUniqName() +". Error: " + e.getMessage());
			}
		}
	}


	public void prepareForCube3()
	{
// 		loadData(); // version is already set to 3.0 ... therefore only
		for (scalasca.cubex.cube.Metric met : metv)
		{
			met.createEnumeration(root_cnodev);
			met.createDataMatrix(cnodev.size(), locv.size());
		}

	}

	static public javax.swing.filechooser.FileFilter getCubeFilesFilter() // returns a fileFilter implementation to be used by java JFileChooser
	{
		return (javax.swing.filechooser.FileFilter)(new scalasca.cubex.cube.services.CubeFileFilter());
	}


}

